"""Plot all spherical harmonic functions for the first 3 el modes."""

import numpy as np
import matplotlib.pyplot as pl
from matplotlib import cm
import matplotlib.gridspec as gridspec

import pyspherical as pysh

spin = 0

lmax = 3

Ntheta = 21
Nphi = 31

theta = np.linspace(0, np.pi, Ntheta)
phi = np.linspace(0, 2 * np.pi, Nphi)

# Meshgrid for function evaluation.
gtheta, gphi = np.meshgrid(theta, phi)

xx = np.sin(gtheta) * np.cos(gphi)
yy = np.sin(gtheta) * np.sin(gphi)
zz = np.cos(gtheta)

fig = pl.figure(figsize=pl.figaspect((lmax + 1) / (2 * lmax + 1)))

gs = gridspec.GridSpec(lmax + 1, 2 * lmax + 1)

for el in range(0, lmax + 1):
    for em in range(-el, el + 1):
        ci = lmax + em
        ax = fig.add_subplot(gs[el, ci], projection='3d')

        sph = pysh.wigner.spin_spherical_harmonic(
            el, em, spin, gtheta, gphi, lmax=lmax)
        cmin, cmax = sph.real.min(), sph.real.max()
        if cmin == cmax:
            sph = 0.5 * np.ones_like(sph, dtype=complex)
        else:
            sph = (sph - cmin) / (cmax - cmin)
        ax.plot_surface(xx, yy, zz, rstride=1, cstride=1,
                        facecolors=cm.plasma(sph.real))
        ax.set_axis_off()
        pos = gs[el, ci].get_position(fig)

        if el == lmax:
            pos = gs[el, ci].get_position(fig)
            fig.text((pos.x0 + pos.x1) / 2, pos.y0, r'$m = $' + str(em))
        if em == 0:
            ax.set_title(r'$\ell = $' + str(el))

pl.show()




import numpy as np
import pytest

import pyspherical as pysh


# Test transforms of sampled data.


@pytest.fixture
def mw_sum_of_harms():
    lmax = 10

    theta, phi = pysh.utils.get_grid_sampling(lmax)

    gtheta, gphi = np.meshgrid(theta, phi)

    def _func(spin=0, lmin=None):
        # Data
        Npeaks = 5

        if lmin is not None:
            lmin = max(spin, lmin) + 1
        else:
            lmin = spin + 1

        # NOTE Transforms seem to fail the loop test when the el = spin component is nonzero.
        peak_els = np.random.choice(np.arange(lmin, lmax - 1), Npeaks, replace=False)
        peak_ems = np.array([np.random.randint(-el, el + 1) for el in peak_els])
        peak_amps = np.random.uniform(10, 20, Npeaks)
        dat = np.zeros(gtheta.shape, dtype=complex)
        for ii in range(Npeaks):
            em = peak_ems[ii]
            el = peak_els[ii]
            dat += peak_amps[ii] * pysh.spin_spherical_harmonic(spin, el, em, gtheta, gphi)

        return dat, lmax, theta, phi, (peak_els, peak_ems, peak_amps)

    return _func


@pytest.mark.parametrize('spin', range(3))
def test_transform_mw_sampling(spin, mw_sum_of_harms):
    # MW sampling:
    #   (lmax) samples in theta
    #   (2 * lmax - 1) in phi

    dat, lmax, theta, phi, (peak_els, peak_ems, peak_amps) = mw_sum_of_harms(spin)
    flm = pysh.forward_transform(dat, phi, theta, lmax, lmin=0, spin=spin)
    Npeaks = len(peak_els)

    # Verify that the peaks are at the expected el, em.
    flm_srt = np.argsort(flm)
    peak_inds = flm_srt[-Npeaks:]
    lmtest = np.array([pysh.ravel_lm(ind) for ind in peak_inds])
    assert set(lmtest[:, 0]) == set(peak_els)
    assert set(lmtest[:, 1]) == set(peak_ems)
    assert np.allclose(np.array(sorted(peak_amps)),
                       flm[peak_inds].real, atol=1e-5)

    # Check that the remaining points are all near zero.
    assert np.allclose(flm[flm_srt[:-Npeaks]], 0.0, atol=1.0)


def test_transform_mw_loop_with_lmin(mw_sum_of_harms):
    # MW sampling:
    #   (lmax) samples in theta
    #   (2 * lmax - 1) in phi

    spin = 0
    lmin = 3
#    pysh.clear_cached_dmat()
    dat, lmax, theta, phi, (peak_els, peak_ems, peak_amps) = mw_sum_of_harms(spin, lmin=lmin)
    flm = pysh.forward_transform(dat, phi, theta, lmax, lmin=lmin, spin=spin)
    Npeaks = len(peak_els)

    # Verify that the peaks are at the expected el, em.
    flm_srt = np.argsort(flm)
    peak_inds = flm_srt[-Npeaks:]
    lmtest = np.array([pysh.ravel_lm(ind + lmin**2) for ind in peak_inds])
    assert set(lmtest[:, 0]) == set(peak_els)
    assert set(lmtest[:, 1]) == set(peak_ems)
    assert np.allclose(np.array(sorted(peak_amps)),
                       flm[peak_inds].real, atol=1e-5)

    # Check that the remaining points are all near zero.
    assert np.allclose(flm[flm_srt[:-Npeaks]], 0.0, atol=1.0)

    # Error with wrong flm shape
    with pytest.raises(ValueError, match='Invalid flm shape'):
        pysh.inverse_transform(flm, phi, theta, lmax, lmin=0, spin=spin)

    # Check that inverse works.
    res = pysh.inverse_transform(flm, phi, theta, lmax, lmin=lmin, spin=spin)

    flm2 = pysh.forward_transform(res, phi, theta, lmax, lmin=lmin, spin=spin)

    assert np.allclose(flm2, flm)
    assert np.allclose(res, dat)


@pytest.mark.parametrize('lmax', range(10, 50, 5))
def test_transform_mw_sampling_monopole(lmax):
    theta, phi = pysh.utils.get_grid_sampling(lmax)

    gtheta, gphi = np.meshgrid(theta, phi)

    amp = 20
    dat = np.ones(gtheta.shape, dtype=float) * amp

    flm = pysh.forward_transform(dat, phi, theta, lmax, lmin=0, spin=0)
    assert np.isclose(flm[0], amp * np.sqrt(4 * np.pi))


@pytest.mark.parametrize('spin', range(3))
def test_transform_mw_sampling_loop(spin, mw_sum_of_harms):
    # sphere -> flm -> sphere.

    dat, lmax, theta, phi, (peak_els, peak_ems, peak_amps) = mw_sum_of_harms(spin)

    flm = pysh.forward_transform(dat, phi, theta, lmax, lmin=0, spin=spin)

    res = pysh.inverse_transform(flm, spin=spin)

    # NOTE High tolerance is needed here.
    # flm and returned data match expectations from pyssht, but do not match the input data.
    # This seems to be highly-dependent on spin and numerical in origin.
    assert np.allclose(res, dat, atol=1)

    Nt = 50
    Nf = 70

    dat2 = pysh.inverse_transform(flm, Nt=Nt, Nf=Nf, spin=spin)

    dth = np.pi / (2 * Nt - 1)
    theta2 = np.linspace(dth, np.pi, Nt, endpoint=True)
    phi2 = np.linspace(0, 2 * np.pi, Nf, endpoint=False)

    flm2 = pysh.forward_transform(dat2, phi2, theta2, lmax, spin=spin)
    res2 = pysh.inverse_transform(flm2, thetas=theta2, phis=phi2, spin=spin)
    assert np.allclose(dat2, res2, atol=1e-4)


@pytest.mark.parametrize('spin', range(3))
def test_loop_mw_nongrid(spin, mw_sum_of_harms):
    # sphere -> flm -> sphere
    # But use meshgrid of points

    dat, lmax, theta, phi, (peak_els, peak_ems, peak_amps) = mw_sum_of_harms(spin)

    gtheta, gphi = np.meshgrid(theta, phi)

    flm = pysh.forward_transform(dat, gphi, gtheta, lmax, spin=spin)
    res = pysh.inverse_transform(flm, gphi, gtheta, lmax, spin=spin)

    assert np.allclose(dat.flatten(), res, atol=1)

# Slow
@pytest.mark.parametrize('offsets', [True, False])
@pytest.mark.parametrize('em', range(1, 15, 4))
def test_loop_diffphi_nongrid(offsets, em):
    # Make a sphere with different phi sampling on each latitude.
    # Evaluate a function on the sphere and do the loop test.

    lmax = 80

    # Number of thetas (latitudes)
    Nt = 81

    # Number of phi samples at each latitude
    Nfmean = 90
    Nf = [Nfmean - 1] * 26 + [Nfmean] * 28 + [Nfmean - 1] * 27

    # Put an offset in phi at each latitude.
    if offsets:
        offsets = [np.random.uniform(0, np.pi / (nf)) for nf in Nf]
    else:
        offsets = np.zeros(Nt)
    thetas, phis = [], []

    dth = np.pi / (2 * Nt - 1)
    base_theta = np.linspace(dth, np.pi, Nt, endpoint=True)

    for ti, th, in enumerate(base_theta):
        nf = Nf[ti]
        thetas.extend([th] * nf)
        cur_phi = (np.linspace(0, 2 * np.pi, nf, endpoint=False) + offsets[ti]) % (2 * np.pi)
        phis.extend(cur_phi.tolist())

    # Evaluate a function on these points.
    el = 17
    em = em
    amp = 50
    dat = amp * pysh.spin_spherical_harmonic(0, el, em, thetas, phis)

    flm = pysh.forward_transform(dat, phis, thetas, lmax=lmax)
    res = pysh.inverse_transform(flm, phis, thetas, lmax=lmax)

    # The results shouldn't match up perfectly, because the sampling isn't exact and
    # it is band-limited, but it's unclear how reversible this should be.
    # This is set with a very high tolerance due to error near the phi = 0.
    tol = 1.0
    assert np.allclose(dat, res, atol=tol)


@pytest.mark.parametrize('spin', range(3))
def test_loop_limited_mem(spin, mw_sum_of_harms):
    # Check that transforms still work with limited memory.

    dat, lmax, theta, phi, (peak_els, peak_ems, peak_amps) = mw_sum_of_harms(spin)

    # NOTE -- Falls into a recursive loop if lmax is too large -- Investigate
    lmax = 60

    flm = pysh.forward_transform(dat, phi, theta, lmax, spin=spin)
    res = pysh.inverse_transform(flm, phi, theta, lmax, spin=spin)
    flm2 = pysh.forward_transform(res, phi, theta, lmax, spin=spin)

    # Error if cached dmat exceeds limit:
    low_limit = 0.025  # MiB

    # Errors, because the current dmat would exceed this.
    with pytest.raises(ValueError, match="Cached DeltaMatrix exceeds"):
        pysh.set_cache_mem_limit(low_limit)

    pysh.clear_cached_dmat()
    # Limit cache memory to a small value.
    pysh.set_cache_mem_limit(low_limit)

    # Test lmin/lmax limiter.
    new_lmin, new_lmax = pysh.wigner.HarmonicFunction._limit_lmin_lmax(0, lmax, True)
    assert new_lmax == lmax and new_lmin > 0
    new_lmin, new_lmax = pysh.wigner.HarmonicFunction._limit_lmin_lmax(0, lmax, False)
    assert new_lmin == 0 and new_lmax < lmax

    # Error if trying to set lmax too high for memory limit.
    with pytest.raises(ValueError, match="Cannot construct DeltaMatrix within"):
        pysh.wigner.HarmonicFunction._set_wigner(0, 500, True)

    # Confirm that transformation still works both ways.
    flm3 = pysh.forward_transform(res, phi, theta, lmax, spin=spin)
    res2 = pysh.inverse_transform(flm3, phi, theta, lmax, spin=spin)

    details = pysh.get_cache_details()

    # Reset cache limit.
    pysh.set_cache_mem_limit(200)

    assert np.allclose(flm3, flm2)
    assert np.allclose(res, res2, atol=1e-4)
    assert details['size'] * np.zeros(1, dtype=np.float32).nbytes < details['cache_mem_limit']



"""
Fast methods for forward and inverse spin-weighted spherical harmonic transforms.

An extension of the methods presented in McEwen and Wiaux (2011) [MW], the transforms here
use FFTs and numba-acceleration to rapidly compute the spin-weighted spherical harmonic
decomposition of a discretely-sampled function on a sphere, or invert the decomposition back to
sampled values. The only requirement is that the sphere be sampled with evenly-spaced samples
in azimuth (phi) at each colatitude (theta).

Non-uniform sampling in theta will be supported in the future.
"""

import numpy as np
from numba import jit

from .wigner import HarmonicFunction, _access_element
from .utils import resize_axis, unravel_lm, get_grid_sampling


__all__ = ['forward_transform', 'inverse_transform']

# -----------------
# Forward transform
# -----------------


def _theta_fft(Gm_th, thetas, lmax, lmin=0, spin=0):

    Nt = thetas.size

    # Apply periodic extension in theta.
    Gm_th = np.pad(Gm_th, [(0, 0), (0, Nt - 1)], mode='constant')
    em = np.fft.ifftshift(np.arange(-(lmax - 1), lmax))
    Gm_th[:, Nt:] = ((-1.0)**(spin + em)
                     * Gm_th[:, 2 * Nt - 2 - np.arange(Nt, 2 * Nt - 1)].T).T

    Fmm = np.fft.fft(Gm_th, axis=1) / (2 * Nt - 1)

    # Truncate/zero-pad the m' axis
    padFmm = resize_axis(Fmm, (2 * lmax - 1), axis=1, mode='zero')

    # Apply phase offset, for thetas with nonzero origin.
    if thetas[0] > 0:
        padFmm = (padFmm * np.exp(-1j * em * np.pi / (2 * Nt - 1)))

    # Convolve with weights over m'

    # Need to shift to zero-centered Fourier ordering for the convolutions.
    padFmm = np.fft.fftshift(padFmm, axes=1)

    def weight(em):
        if np.abs(em) == 1:
            return np.sign(em) * np.pi * 1j / 2.
        if em % 2 == 1:
            return 0
        return 2 / (1 - em**2)

    em_ext = range(-2 * (lmax - 1), 2 * (lmax - 1) + 1)
    ite = (weight(mm) for mm in em_ext)
    weights = np.fromiter(ite, dtype=complex)[::-1]

    def do_conv(a):
        # The "valid" option only works with these Fourier-transformed
        # quantities if they have been FFT-shifted, such that the 0 mode
        # is in the center of the array.
        return np.convolve(a, weights, mode='valid')

    Gmm = np.apply_along_axis(do_conv, 1, padFmm) * 2 * np.pi

    # Unshift the m' axis
    Gmm = np.fft.ifftshift(Gmm, axes=1)

    return Gmm


@jit(nopython=True)
def _jit_dmm2flm(dmm, cur_lmax, spin, dmatarr, lmin, flm, out_lmin):
    # Faster evaluation.
    # flm = output array, written to directly.

    loopmin = max(out_lmin, spin, lmin)
    for el in range(loopmin, cur_lmax):
        prefac = np.sqrt((2 * el + 1) / (4 * np.pi))
        for m in range(-el, el + 1):
            ind = unravel_lm(el, m) - out_lmin**2
            # The MW paper was missing a factor of (-1)**(m+spin) here.
            prefac2 = (-1)**(m + 2 * spin) * (1j)**(m + spin)
            # mp == 0
            flm[ind] += _access_element(el, 0, m, dmatarr, lmin=lmin) \
                * _access_element(el, 0, -spin, dmatarr, lmin=lmin) * dmm[m, 0]
            negfac = (-1)**(m + spin)
            for mp in range(-el, 0):
                dels = _access_element(el, mp, m, dmatarr, lmin=lmin) \
                    * _access_element(el, mp, -spin, dmatarr, lmin=lmin)
                flm[ind] += dels * (dmm[m, mp] + negfac * dmm[m, -mp])
            flm[ind] *= prefac * prefac2


def _dmm_to_flm(dmm, lmax, spin, flm=None, lmin=0, out_lmin=0):
    # Recursion is to handle potential memory-limited situations:
    #  - If the cached Wigner matrix has an lmax < less than given lmax,
    #    then rerun with the smaller lmax and then with the remainder.
    #  - Otherwise, fill flm using the JIT-compiled function above.
    if flm is None:
        out_lmin = lmin
        flm = np.zeros(lmax**2 - out_lmin**2, dtype=complex)
    HarmonicFunction._set_wigner(lmin, lmax, high=False)
    wig_d = HarmonicFunction.current_dmat
    if wig_d.lmax < lmax:
        _dmm_to_flm(dmm, wig_d.lmax, spin, flm=flm, lmin=lmin)
        _dmm_to_flm(dmm, lmax, spin, flm=flm, lmin=wig_d.lmax)
    else:
        _jit_dmm2flm(dmm, lmax, spin, wig_d._arr, wig_d.lmin, flm, out_lmin)

    return flm


def _do_fwd_transform_on_grid(dat, phis, thetas, lmax, lmin=0, ph_ax=0, th_ax=1, spin=0):
    """Do forward transform, assuming a regular grid in both theta and phi."""
    if th_ax == 0:
        # Underlying functions expect the theta and phi axes to be 1 and 0, resp.
        dat = dat.T

    Nf, Nt = dat.shape

    # Transform phi to m and pad/truncate.
    dm_th = np.fft.fft(dat, axis=0) / Nf
    dm_th = resize_axis(dm_th, (2 * lmax - 1), mode='zero', axis=0)

    # Transform theta to m'.
    # If evenly-spaced in theta, can use an FFT.

    dth = np.diff(thetas)
    if np.allclose(dth, dth[0]):
        dmm = _theta_fft(dm_th, thetas, lmax, lmin, spin)
    else:
        raise NotImplementedError(
            "Non-uniform latitude spacing is not yet supported.")
    flm = _dmm_to_flm(dmm, lmax, spin, lmin=lmin)

    return flm


def _do_fwd_transform_nongrid(dat, phis, thetas, lmax, lmin, spin):
    """
    Do forward transform without assuming the same phi sampling for all colatitudes.

    Assumes uniformly-spaced samples in phi on each colatitude ring.
    """
    dat = dat.flatten()
    phis = phis.flatten()
    thetas = thetas.flatten()

    un_thetas, lat_ind = np.unique(thetas, return_inverse=True)
    Nlats = un_thetas.size

    # phi to m, per ring
    dm_th = np.zeros((2 * lmax - 1, Nlats), dtype=complex)

    em = np.fft.ifftshift(np.arange(-(lmax - 1), lmax))
    for th_i, th in enumerate(un_thetas):
        ring = lat_ind == th_i
        dat_i = dat[ring]
        Nf = dat_i.size
        dm_th[:, th_i] = resize_axis(np.fft.fft(
            dat_i) / Nf, (2 * lmax - 1), mode='zero')
        # Apply phase offset when the ring doesn't start at phi = 0
        phase = np.exp(-1j * em * np.min(np.abs(phis[ring])))
        dm_th[:, th_i] *= phase

    # theta to m'
    dth = np.diff(un_thetas)
    if np.allclose(dth, dth[0]):
        dmm = _theta_fft(dm_th, un_thetas, lmax, lmin, spin)
    else:
        raise NotImplementedError(
            "Non-uniform latitude spacing is not yet supported.")

    flm = _dmm_to_flm(dmm, lmax, spin, lmin=lmin)

    return flm


def forward_transform(dat, phis, thetas, lmax, lmin=0, spin=0):
    """
    Transform sampled data to spin-weighted spherical harmonics.

    Parameters
    ----------
    dat: ndarray of float or complex
        Sampled data array.
        Shape can be:
            * dat.shape = (phis.size, thetas.size)
            * dat.shape = (thetas.size, phis.size)
            * dat.shape = phis.shape = theta.shape
        In the first two cases, it is assumed that the same
        sampling in phi is done at all co-latitudes.
    phis: ndarray of float
        Sample azimuths in radians.
    thetas: ndarray of float
        Sample co-latitudes in radians.
    lmax: int
        Maximum multipole mode.
    lmin: int
        Minimum multipole mode to compute. (Optional, default 0)
    spin: int
        Spin of the transform.

    Returns
    -------
    flm: ndarray of complex
        Spherical-harmonic components, shape ((lmax-1)**2)
        Ordered as:
            el  = 0  1  1  1  2  2  2  2  2
            em  = 0 -1  0  1 -2 -1  0  1  2
            ind = 0  1  2  3  4  5  6  7  8
        Modes with el < spin will be set to zero.
    """

    phis = np.asarray(phis)
    thetas = np.asarray(thetas)
    grid = True
    if dat.shape == (phis.size, thetas.size):
        fax, tax = 0, 1
    elif dat.shape == (thetas.size, phis.size):
        fax, tax = 1, 0
    elif dat.shape == thetas.shape == phis.shape:
        grid = False
    else:
        raise ValueError("Data shapes inconsistent:"
                         "\n\tdat.shape = {}"
                         "\n\tphis.shape = {}"
                         "\n\tthetas.shape = {}".format(
                             str(dat.shape), str(phis.shape), str(thetas.shape))
                         )

    if grid:
        return _do_fwd_transform_on_grid(dat, phis, thetas, lmax, lmin, fax, tax, spin)

    else:
        return _do_fwd_transform_nongrid(dat, phis, thetas, lmax, lmin, spin)


# -----------------
# Inverse transform
# -----------------

@jit(nopython=True)
def _jit_flm2fmm(flm, cur_lmax, spin, dmatarr, lmin, fmm, in_lmin):
    loopmin = max(lmin, in_lmin, spin)
    for el in range(loopmin, cur_lmax):
        prefac = (-1)**spin * np.sqrt((2 * el + 1) / (4 * np.pi))
        for m in range(-el, el + 1):
            prefac2 = (1j)**(-m - spin) * flm[unravel_lm(el, m) - in_lmin**2]
            # mp == 0
            fmm[m, 0] += prefac * prefac2 * _access_element(el, 0, m, dmatarr, lmin=lmin) \
                * _access_element(el, 0, -spin, dmatarr, lmin=lmin)
            negfac = (-1)**(m + spin)
            for mp in range(1, el + 1):
                val = prefac * prefac2 * _access_element(el, m, mp, dmatarr, lmin=lmin) \
                    * _access_element(el, -spin, mp, dmatarr, lmin=lmin)
                fmm[m, mp] += val
                fmm[m, -mp] += val * negfac


def _flm_to_fmm(flm, lmax, spin, fmm=None, lmin=0, in_lmin=0):
    if fmm is None:
        in_lmin = lmin
        fmm = np.zeros(((2 * lmax - 1), (2 * lmax - 1)), dtype=complex)
    HarmonicFunction._set_wigner(lmin, lmax, high=False)
    wig_d = HarmonicFunction.current_dmat
    if wig_d.lmax < lmax:
        _flm_to_fmm(flm, wig_d.lmax, spin, fmm=fmm, lmin=lmin, in_lmin=in_lmin)
        _flm_to_fmm(flm, lmax, spin, fmm=fmm, lmin=wig_d.lmax, in_lmin=in_lmin)
    else:
        _jit_flm2fmm(flm, lmax, spin, wig_d._arr, wig_d.lmin, fmm, in_lmin)

    return fmm


def _theta_ifft(fmm, lmax, spin, Nt=None, offset=None):
    # offset = np.pi / (2 * Nt - 1) for MW sampling

    em = np.fft.ifftshift(np.arange(-(lmax - 1), lmax))

    # Apply phase offset
    fmm *= np.exp(1j * em * offset)

    # Apply ifft over theta
    fmm = resize_axis(fmm, 2 * Nt - 1, axis=1)
    Fm_th = np.fft.ifft(fmm, axis=1) * (2 * Nt - 1)

    # Cut the periodic extension in theta
    Fm_th = Fm_th[:, :Nt]

    return Fm_th


def _do_inv_transform_on_grid(flm, phis, thetas, lmax, lmin=0, spin=0):
    """Do inverse transform assuming regular grid in theta and phi."""
    Fmm = _flm_to_fmm(flm, lmax, spin, lmin=lmin)

    # Transform over theta
    theta_is_equi = np.allclose(
        thetas[1] - thetas[0], thetas[2:] - thetas[1:-1])

    if theta_is_equi:
        Fm_th = _theta_ifft(
            Fmm, lmax, spin, offset=thetas[0], Nt=thetas.size)
    else:
        raise NotImplementedError(
            "Non-uniform latitude spacing is not yet supported.")

    # Transform over phi
    Nf = phis.size
    Fm_th = resize_axis(Fm_th, Nf, axis=0)
    dat = np.fft.ifft(Fm_th, axis=0) * Nf

    return dat


def _do_inv_transform_nongrid(flm, phis, thetas, lmax, lmin=0, spin=0):
    """
    Do inverse transform without assuming the same phi sampling for all colatitudes.

    Assumes uniformly-spaced samples in phi on each colatitude ring.
    """
    phis = phis.flatten()
    thetas = thetas.flatten()

    Fmm = _flm_to_fmm(flm, lmax, spin, lmin=lmin)

    un_thetas, lat_ind = np.unique(thetas, return_inverse=True)
    Nlats = un_thetas.size

    # Transform over theta
    theta_is_equi = np.allclose(
        un_thetas[1] - un_thetas[0], un_thetas[2:] - un_thetas[1:-1])

    if theta_is_equi:
        Fm_th = _theta_ifft(
            Fmm, lmax, spin, offset=un_thetas[0], Nt=Nlats
        )
    else:
        raise NotImplementedError(
            "Non-uniform latitude spacing is not yet supported.")

    # Transform m to phi, per ring.
    #   Need to apply a phase offset if the 0th index is not at phi = 0
    em = np.fft.ifftshift(np.arange(-(lmax - 1), lmax))
    dat = np.zeros(phis.size, dtype=complex)
    for th_i, th in enumerate(un_thetas):
        ring = lat_ind == th_i
        phi_i = np.sort(phis[ring])
        Nf = phi_i.size
        phase = np.exp(1j * em * phi_i[0])
        ring_dat = phase * Fm_th[:, th_i]
        dat[ring] = np.fft.ifft(resize_axis(ring_dat, Nf)) * Nf

    return dat


def inverse_transform(flm, phis=None, thetas=None, lmax=None, lmin=0, spin=0, Nt=None, Nf=None):
    """
    Evaluate spin-weighted spherical harmonic components to a sphere.

    Without any optional parameters, this assumes the sampling described in McEwan and Wiaux (2011).

    Parameters
    ----------
    flm: ndarray of complex
        Spherical harmonic components in the ordering returned by forward_transform.
    phis: ndarray of float
        Sample azimuths in radians. Overrides Nf keyword. (Optional)
        Defaults to Nf uniform samples in [0,  2 pi).
    thetas: ndarray of float
        Sample co-latitudes in radians. Overrides Nt keyword. (Optional)
        Defaults to Nt uniform samples in [dth, pi], where dth = pi/(2 * Nt - 1)
    lmax: int
        Maximum multipole mode. (Optional)
        Defaults to sqrt(len(flm)).
    lmin: int
        Minimum multipole mode to compute. (Optional, default 0)
    spin: int
        Spin of the transform.
    Nt: int
        Number of samples in theta. (Optional, defaults to lmax)
    Nf: int
        Number of samples in phi for all colatitudes (Optional, defaults to (2 * lmax - 1))

    Returns
    -------
    dat: ndarray of complex
        Evaluation of the spherical harmonic components to the sample positions.


    Notes
    -----
    The sampling given through phi/theta can either be the axes of a grid
    (i.e., identical phi sampling for all thetas) or specify the individual phi/theta positions
    for each sample. In the first case, dat.shape = (phi.size, theta.size). In the latter case,
    dat.shape = phi.shape = theta.shape.

    This function cannot be used to evaluate the spherical harmonic components to arbitrary
    positions on the sphere. For that, the function
    :meth:`HarmonicFunction.spin_spherical_harmonic()` is provided.
    """
    if lmax is None:
        lmax = int(np.sqrt(flm.size))

    if np.sqrt(flm.size + lmin**2) % 1 != 0:
        raise ValueError("Invalid flm shape: {}".format(str(flm.shape)))

    _thetas, _phis = get_grid_sampling(lmax, Nt, Nf)
    if thetas is None:
        thetas = _thetas
    if phis is None:
        phis = _phis

    phis = np.asarray(phis)
    thetas = np.asarray(thetas)

    Nt = thetas.size
    Nf = phis.size

    # Guess from coordinate shapes/sizes if they represent axes of a grid or
    # fully cover the desired sampling.
    grid = True
    if not np.unique(thetas).size == thetas.size:
        grid = False
        if not phis.shape == thetas.shape:
            raise ValueError("It looks like the coordinates are non-grid, "
                             "but their shapes are inconsistent:"
                             "\n\tphis.shape = {}"
                             "\n\tthetas.shape = {}".format(str(phis.shape), str(thetas.shape)))

    if grid:
        return _do_inv_transform_on_grid(flm, phis, thetas, lmax, lmin, spin)

    else:
        return _do_inv_transform_nongrid(flm, phis, thetas, lmax, lmin, spin)



"""
An example of using spherical harmonic transforms and evaluation.

Generate a linear combination of spherical harmonics and
plot the transform, showing expected peaks.

"""


import numpy as np
import matplotlib.pyplot as pl

import pyspherical as pysh

lmax = 10
spin = 1

Npeaks = 7

Ntheta = 21
Nphi = 31

# Isolatitude and isolongitude spherical sampling.
theta, phi = pysh.utils.get_grid_sampling(Nt=Ntheta, Nf=Nphi)

# Meshgrid for function evaluation.
gtheta, gphi = np.meshgrid(theta, phi)

# Choose (el, em) harmonic modes to include.
peak_els = np.random.choice(np.arange(spin + 1, lmax - 1), Npeaks, replace=False)
peak_ems = np.array([np.random.randint(-el, el + 1) for el in peak_els])
peak_amps = np.random.uniform(10, 20, Npeaks)
dat = np.zeros(gtheta.shape, dtype=complex)
for ii in range(Npeaks):
    em = peak_ems[ii]
    el = peak_els[ii]
    dat += peak_amps[ii]\
        * pysh.spin_spherical_harmonic(el, em, spin, gtheta, gphi, lmax=lmax)

# Perform forward transform from data to harmonic components.

flm = pysh.forward_transform(dat, phi, theta, lmax, spin=spin)

# Plot the data, the flm, and the expected peak locations.
fig, (ax0, ax1) = pl.subplots(1, 2)

im = ax0.pcolormesh(np.degrees(theta), np.degrees(phi), dat.real)
pl.colorbar(im, ax=ax0)

ax1.plot(flm.real, label='Result')

# Get index in flm array from (el, em)
indices = [pysh.unravel_lm(peak_els[ii], peak_ems[ii]) for ii in range(Npeaks)]
ax1.scatter(indices, peak_amps, marker='o', color='r', label='Expected')

fig.suptitle(f"Spin {spin}", fontsize=14)
ax0.set_title("Re[Data]")
ax0.set_xlabel(r'$\theta$ [deg]')
ax0.set_ylabel(r'$\phi$ [deg]')

ax1.set_title(r"$_sf_{lm}$")
ax1.set_xlabel(r'index')
ax1.set_ylabel(r'Amplitude')
ax1.legend()
pl.tight_layout()
pl.show()




import numpy as np
import pytest

import pyspherical as pysh


def test_get_mw_grid():
    # Check that different methods fo getting an MW sampling grid are consistent and correct.
    lmax = 10
    Nt = lmax
    Nf = 2 * lmax - 1

    with pytest.raises(ValueError, match="Need to provide lmax"):
        th, ph = pysh.utils.get_grid_sampling()

    th, ph = pysh.utils.get_grid_sampling(lmax)
    assert th.size == Nt
    assert ph.size == Nf


@pytest.mark.parametrize('Nitems', [10, 11])
@pytest.mark.parametrize('Nnew', [5, 6, 10, 13, 14])
def test_resize_axis(Nitems, Nnew):
    # Zero-pad / truncate of arrays.
    #   Check that:
    #       * Data end up in the right places, with nothing lost, when zero-padding.
    #       * The right parts of the array are kept when truncating.

    arr0 = np.arange(Nitems) + 1
    arr1 = pysh.utils.resize_axis(arr0, Nnew, mode='zero', axis=0)
    arr2 = pysh.utils.resize_axis(arr0, Nnew, mode='start', axis=0)
    arr3 = pysh.utils.resize_axis(arr0, Nnew, mode='center', axis=0)

    newodd = Nnew % 2 == 1
    limit = min(Nnew, Nitems) // 2

    # "start" mode -- either truncate to the start of the array, or
    # pad with zeros after arr0.
    assert (arr2[:limit] == arr0[:limit]).all()
    if newodd:
        # "zero" mode -- array should match original, with zeros in the middle.
        assert (arr0[:limit] == arr1[:limit]).all()
        assert (arr0[-limit + 1:] == arr1[-limit + 1:]).all()
    else:
        assert (arr0[:limit] == arr1[:limit]).all()
        assert (arr0[-limit:] == arr1[-limit:]).all()

    diff = Nnew - Nitems
    if diff < 0:
        trunced = arr0.copy().tolist()
        for ii in range(-diff):
            if ii % 2 == 1:
                del trunced[0]
            else:
                del trunced[-1]
        assert trunced == arr3.tolist()
    else:
        padded = arr3.copy().tolist()
        for ii in range(diff):
            if ii % 2 == 1:
                del padded[0]
            else:
                del padded[-1]
        assert padded == arr0.tolist()

    # Now check that the sizes are correct and the original data is kept
    # in the right order with zero-padding.
    for arr in [arr1, arr2, arr3]:
        assert arr.size == Nnew
        if Nnew >= Nitems:
            assert (arr[arr > 0] == arr0).all()



import numpy as np
import pytest

import pyspherical as pysh


@pytest.fixture
def mw_sampling():
    # MW sampling of a sphere
    Nt = 701   # Number of samples in theta (must be odd)
    Nf = 401  # Samples in phi
    theta, phi = pysh.utils.get_grid_sampling(Nt=Nt, Nf=Nf)
    gtheta, gphi = np.meshgrid(theta, phi)
    return theta, phi, gtheta, gphi


@pytest.mark.parametrize('slm', ((spin, el, em)
                                 for spin in range(3)
                                 for el in range(spin, 5)
                                 for em in range(-el, el + 1)
                                 )
                         )
def test_transform_eval_compare(mw_sampling, slm):
    # Compare evaluation of the spherical harmonic to the result returned
    # by the inverse transform.
    # Also compare spin_spherical_harmonic to goldberg eval.
    amp = 20
    lmax = 5
    spin, el, em = slm

    theta, phi, gtheta, gphi = mw_sampling

    if (spin, el, em) == (0, 0, 0):
        pysh.clear_cached_dmat()   # Reset current_dmat
        assert pysh.get_cached_dmat() is None

    flm = np.zeros(lmax**2, dtype=complex)
    flm[pysh.utils.unravel_lm(el, em)] = amp

    # Error cases
    with pytest.raises(ValueError, match="theta and phi must have"):
        pysh.spin_spherical_harmonic(spin, el, em, theta, gphi)

    with pytest.raises(ValueError, match="theta and phi must have"):
        pysh.spin_spharm_goldberg(spin, el, em, theta, gphi)

    test1 = pysh.inverse_transform(flm, thetas=theta, phis=phi, spin=spin)
    test2 = amp * \
        pysh.spin_spherical_harmonic(
            spin, el, em, gtheta, gphi, lmax=lmax)
    test3 = amp * pysh.wigner.spin_spharm_goldberg(spin, el, em, gtheta, gphi)

    assert np.allclose(test1, test2, atol=1e-10)
    assert np.allclose(test2, test3, atol=1e-5)
    assert pysh.get_cached_dmat().lmax == lmax


def test_wigner_symm():
    # Test symmetries of the Wigner-d function.
    # Appendix of Prezeau and Reinecke (2010)

    # Test error case:
    with pytest.raises(ValueError, match='el 3 is less than lmin'):
        pysh.wigner_d(3, 0, 0, 0.5, lmin=5)

    lmax = 15

    def dl(el, m1, m2, theta):
        return pysh.wigner_d(el, m1, m2, theta, lmax=lmax)

    Nth = 5
    for th in np.linspace(0, np.pi, Nth):
        for ll in range(lmax):
            for m in range(-ll, ll + 1):
                for mm in range(-ll, ll + 1):
                    val1 = dl(ll, m, mm, th)
                    assert np.isclose(val1, (-1)**(m - mm) * dl(ll, -m, -mm, th))
                    assert np.isclose(val1, (-1)**(m - mm) * dl(ll, mm, m, th))
                    assert np.isclose(val1, dl(ll, -mm, -m, th))
                    assert np.isclose(dl(ll, m, mm, -th),
                                      dl(ll, mm, m, th), atol=1e-10)
                    assert np.isclose(
                        dl(ll, m, mm, np.pi - th), (-1)**(ll - mm) * dl(ll, -m, mm, th), atol=1e-10)


def test_delta_matrix_inits():
    # Initializing delta matrices with different conditions.

    lmax = 20

    # Full:
    dmat1 = pysh.DeltaMatrix(lmax)

    # With lmin, but no starting array.
    dmat2 = pysh.DeltaMatrix(lmax, lmin=5)

    # Subset of existing arr0:
    dmat3 = pysh.DeltaMatrix(15, lmin=6, dmat=dmat1)

    # Starting from end of previous.
    dmat4 = pysh.DeltaMatrix(lmax, lmin=15, dmat=dmat3)

    # Starting from after end of previous.
    dmat5 = pysh.DeltaMatrix(20, lmin=9, dmat=pysh.DeltaMatrix(7))

    # Given an existing dmat, will it be copied correctly?
    dmat6 = pysh.DeltaMatrix(lmax, lmin=0, dmat=dmat1)

    # For each, check that the results match with the full.

    for dm in [dmat2, dmat3, dmat4, dmat5, dmat6]:
        ln, lx = dm.lmin, dm.lmax
        for el in range(ln, lx):
            for m1 in range(-el, el + 1):
                for m2 in range(-el, el + 1):
                    assert dm[el, m1, m2] == dmat1[el, m1, m2]

    # Check error cases:
    with pytest.raises(ValueError, match="l < lmin. Need to re-evaluate"):
        dmat2[3, 0, 1]
    with pytest.raises(ValueError, match="l > lmax. Need to re-evaluate"):
        dmat2[22, 0, 1]


@pytest.mark.parametrize(
    ('err', 'lmin', 'lmax', 'arrsize'),
    [
        (False, 0, 15, 816),
        (False, 4, 9, 200),
        (True, 4, 9, 157)
    ]
)
def test_dmat_params(err, lmin, lmax, arrsize):
    # Getting missing parameter from the set lmin, lmax, arrsize
    _get_array_params = pysh.DeltaMatrix._get_array_params
    if not err:
        assert _get_array_params(None, lmax, arrsize) == (lmin, lmax, arrsize)
        assert _get_array_params(lmin, None, arrsize) == (lmin, lmax, arrsize)
        assert _get_array_params(lmin, lmax, None) == (lmin, lmax, arrsize)
    else:
        with pytest.raises(ValueError, match="Invalid"):
            _get_array_params(None, lmax, arrsize, strict=True)
        with pytest.raises(ValueError, match="Invalid"):
            _get_array_params(lmin, None, arrsize, strict=True)


@pytest.mark.parametrize(
    ('lmin', 'lmax', 'arrsize'),
    [
        (0, 15, 400),
        (4, 9, 70),
        (4, 9, 157)
    ]
)
def test_dmat_params_nonstrict(lmin, lmax, arrsize):
    # Ensure the returned size is smaller than the limits for various lmin/lmax.
    _get_array_params = pysh.DeltaMatrix._get_array_params

    # Choose a new lmin:
    lmin_new, _, arrsize_new = _get_array_params(lmax=lmax, arrsize=arrsize)
    assert (lmin_new >= lmin) and (arrsize_new <= arrsize)

    # Choose a new lmax:
    _, lmax_new, arrsize_new = _get_array_params(lmin=lmin, arrsize=arrsize)
    assert (lmax_new <= lmax) and (arrsize_new <= arrsize)


# Slow
def test_dmat_largeL():
    # Check that DeltaMatrix elements are accurate for some large el,
    # compared with the precise results of sympy.
    # This test may be slow.

    pytest.importorskip('sympy')
    from sympy.physics.quantum.spin import Rotation
    el, m, mp = 200, 77, 59

    wigd = Rotation.d(el, m, mp, np.pi / 2)
    val = complex(wigd.doit())

    pysh.wigner.HarmonicFunction._set_wigner(180, 210)

    assert np.isclose(val.real, pysh.wigner_d(el, m, mp, np.pi / 2), atol=1e-6)


# Slow
def test_dmat_64bit():
    # Check the DeltaMatrix calculated with double precision.
    # TODO Add option to run through the full range of m/mp, when allowing slow tests.
    pytest.importorskip('sympy')
    from sympy.physics.quantum.spin import Rotation
    el = 200

    pysh.wigner.HarmonicFunction._set_wigner(el - 1, el + 1, dtype=np.float64)

    # Choose random m, mp.
    for m, mp in np.random.randint(0, el + 1, (1, 2)):
        wigd = Rotation.d(el, m, mp, np.pi / 2)
        val = complex(wigd.doit())

        # Different sign convention.
        assert np.isclose(val.real, (-1)**(m + mp)
                          * pysh.wigner_d(el, m, mp, np.pi / 2), atol=1e-12)


def test_error():
    with pytest.raises(Exception, match="HarmonicFunction class is not"):
        pysh.wigner.HarmonicFunction()



from setuptools import setup
import io

with io.open('README.md', 'r', encoding='utf-8') as readme_file:
    readme = readme_file.read()

setup_args = {
    'name': 'pyspherical',
    'author': 'Adam E. Lanman',
    'url': 'https://github.com/aelanman/pyspherical',
    'download_url': 'https://github.com/aelanman/pyspherical/archive/v0.0.1.tar.gz',
    'license': 'MIT',
    'description': '',
    'long_description': readme,
    'long_description_content_type': 'text/markdown',
    'package_dir': {'pyspherical': 'pyspherical'},
    'packages': ['pyspherical', 'pyspherical.tests'],
    'version': '0.0.1',
    'include_package_data': True,
    'test_suite': 'pytest',
    'tests_require': ['pytest', 'sympy'],
    'setup_requires': ['pytest-runner'],
    'install_requires': ['numpy', 'numba', 'scipy'],
    'classifiers': ['Development Status :: 3 - Alpha',
                    'Intended Audience :: Science/Research',
                    'License :: OSI Approved :: MIT License',
                    'Programming Language :: Python :: 3.6',
                    'Topic :: Scientific/Engineering :: Physics'],
    'keywords': 'spin spherical harmonic transforms wigner'
}

if __name__ == '__main__':
    setup(**setup_args)



"""Utility functions."""

import numpy as np

from numba import jit, int32, types

__all__ = ['resize_axis', 'tri_ravel', 'unravel_lm', 'ravel_lm', 'get_grid_sampling']


def resize_axis(arr, size, mode='zero', axis=0):
    """
    Resize an axis of the array, either truncating or zero-padding.

    The "mode" keyword determines how this is done.

    Parameters
    ----------
    arr: ndarray
        Array to resize
    size: int
        New size for the axis.
    mode: str
        If the new array is zero-padded, where to put the old data:
            * 'zero' : Put zeros in the middle of the axis (center data on zero)
            * 'start' : Put zeros at the end of the axis.
            * 'center': Evenly fill data on both sides.
        Defaults to "zero".
    axis: int
        Which axis to resize

    Returns
    -------
    arr: ndarray
        Input array with the specified axis padded or truncated.

    Notes
    -----
    The 'zero' mode can be used to zero-pad an FFT-transformed array before
    applying an inverse transform.
    """
    shape = list(arr.shape)
    shape[axis] = size

    new = np.zeros(tuple(shape), dtype=arr.dtype)

    oldodd = arr.shape[axis] % 2 == 1
    newodd = size % 2 == 1

    L = arr.shape[axis]
    if oldodd:
        center = (L - 1) // 2   # This stays on the left.
    else:
        center = L // 2
    if newodd:
        newcent = (size - 1) // 2
    else:
        newcent = size // 2

    _arr = np.swapaxes(arr, axis, 0)
    _new = np.swapaxes(new, axis, 0)

    if mode == 'zero':
        limit = np.min([center, newcent])
        if oldodd:
            _new[:limit + 1, ...] = _arr[:limit + 1, ...]
            _new[-1:-limit - 1:-1, ...] = _arr[-1:-limit - 1:-1, ...]
        else:
            _new[:limit, ...] = _arr[:limit, ...]
            _new[-1:-limit - 1:-1, ...] = _arr[-1:-limit - 1:-1, ...]

    elif mode == 'start':
        limit = np.min([size, L])
        _new[:limit, ...] = _arr[:limit, ...]
    elif mode == 'center':
        if size <= L:
            # Truncating
            base = int(np.floor((L - size) / 2))
            _new[:, ...] = _arr[base:base + size, ...]
        else:
            base = int(np.floor((size - L) / 2))
            _new[base:base + L, ...] = _arr[:, ...]

    new = np.swapaxes(_new, 0, axis)

    return new


# Index raveling/unraveling

@jit(int32(int32, int32, int32), nopython=True)
def tri_ravel(l, m1, m2):
    """Ravel indices for the 'stack of triangles' ordering."""
    # m1 must be >= m2
    if m1 < m2 or m1 > l or m2 > l or m1 < 0 or m2 < 0:
        raise ValueError("Invalid indices")
    base = l * (l + 1) * (l + 2) // 6
    offset = (l - m1) * (l + 3 + m1) // 2 + m2
    ind = base + offset
    return int(ind)


@jit(int32(int32), nopython=True)
def tri_base(l):
    """Minimum index for a given el block."""
    return tri_ravel(l, l, 0)


@jit(int32(int32), nopython=True)
def el_block_size(l):
    """Size needed for a given el in the dmat array."""
    return (l + 1) * (l + 2) // 2


@jit(int32(int32, int32), nopython=True)
def unravel_lm(el, m):
    """Get index from (el, em)."""
    return el * (el + 1) + m


@jit(types.Tuple((int32, int32))(int32), nopython=True)
def ravel_lm(ind):
    """Get (el, em) from index."""
    el = int(np.floor(np.sqrt(ind)))
    m = ind - el * (el + 1)
    return el, m


def get_grid_sampling(lmax=None, Nt=None, Nf=None):
    """
    Get sample positions for "grid", compatible with the MW transform methods.

    Parameters
    ----------
    lmax: int
        Maximum multipole moment (Optional if Nt/Nf are set.)
    Nt: int
        Number of samples in theta. (Optional, defaults to lmax)
    Nf: int
        Number of samples in phi for all colatitudes
        (Optional, defaults to (2 * lmax - 1))

    Returns
    -------
    thetas: ndarray
        Colatitudes in radians.
    phis: ndarray
        Azimuths in radians.
    """
    if (lmax is None) and (Nt is None or Nf is None):
        raise ValueError("Need to provide lmax if Nt and Nf are unset.")

    if Nt is None:
        Nt = lmax

    if Nf is None:
        if lmax is not None:
            Nf = 2 * lmax - 1

    dth = np.pi / (2 * Nt - 1)
    thetas = np.linspace(dth, np.pi, Nt, endpoint=True)
    phis = np.linspace(0, 2 * np.pi, Nf, endpoint=False)

    return thetas, phis



"""Evaluation of Wigner-d functions and spin-weighted spherical harmonic functions."""

import numpy as np
import warnings
from numba import jit
from scipy.special import binom, factorial

from .utils import tri_ravel, tri_base, el_block_size


__all__ = [
    'spin_spharm_goldberg',
    'DeltaMatrix',
    'spin_spherical_harmonic',
    'wigner_d',
    'get_cached_dmat',
    'clear_cached_dmat',
    'set_cache_mem_limit',
    'get_cache_details',
]


@jit(nopython=True)
def _dmat_eval(lmax, arr, lmin=0, lstart=None, arr0=None):
    # Evaluate the values of the Wigner-d matrices at pi/2.
    # arr = linear array, modified in-place
    if arr0 is not None:
        arr[:len(arr0)] = arr0
    else:
        arr[0] = 1.0
    if lstart is None:
        lstart = lmin

    offset = tri_ravel(lmin, lmin, 0)
    for el in range(lstart + 1, lmax + 1):
        if el <= lmin + 1:
            # Shift previous result back.
            elm2_size = el_block_size(el - 2)
            # If this is the first step, the el-1 block is
            # in the zeroth position.
            if el == lstart + 1:
                elm2_size = 0
            elm1_size = el_block_size(el - 1)
            arr[:elm1_size] = arr[elm2_size:elm2_size + elm1_size].copy()
            offset = tri_base(el - 1)
        dll0 = -np.sqrt((2 * el - 1) / float(2 * el)) * \
            arr[tri_ravel(el - 1, el - 1, 0) - offset]
        arr[tri_ravel(el, el, 0) - offset] = dll0
        for m2 in range(0, el + 1):
            if m2 > 0:
                dllm = np.sqrt(
                    (el / 2) * (2 * el - 1) / ((el + m2) * (el + m2 - 1))
                ) * arr[tri_ravel(el - 1, el - 1, m2 - 1) - offset]
                arr[tri_ravel(el, el, m2) - offset] = dllm
            for m1 in range(el - 1, m2 - 1, -1):
                fac1 = (2 * m2) / np.sqrt((el - m1) * (el + m1 + 1)) * \
                    arr[tri_ravel(el, m1 + 1, m2) - offset]
                fac2 = 0.0
                if (m1 + 2) <= el:
                    fac2 = np.sqrt(
                        ((el - m1 - 1) * (el + m1 + 2))
                        / ((el - m1) * (el + m1 + 1))
                    ) * arr[tri_ravel(el, m1 + 2, m2) - offset]
                arr[tri_ravel(el, m1, m2) - offset] = fac1 - fac2


@jit(nopython=True)
def _access_element(l, m1, m2, arr, lmin=0):
    # Access stored elements, or use
    # symmetry relations for non-stored elements.

    _m1, _m2 = m1, m2

    fac = (-1.)**(_m2 - _m1)     # For sign convention of the SSHT paper

    if _m1 < 0 and _m2 < 0:
        fac *= (-1)**(m1 - m2)
        m1 *= -1
        m2 *= -1
    elif _m1 < 0 and _m2 >= 0:
        fac *= (-1)**(l - _m2)
        m1 *= -1
    elif _m1 >= 0 and _m2 < 0:
        fac *= (-1)**(l + _m1)
        m2 *= -1

    # reverse if wrong order
    if m1 < m2:
        fac *= (-1.)**(m1 - m2)
        m1, m2 = m2, m1

    val = fac * arr[tri_ravel(l, m1, m2) - tri_ravel(lmin, lmin, 0)]
    if np.isnan(val):
        raise ValueError("Invalid entry in dmat")
    return val


class DeltaMatrix:
    """
    Wigner-d functions evaluated at pi/2.

    Only stores values for m1, m2 >= 0. Other values are returned by symmetry relations.

    Based on the methods in:
        S, Trapani, and Navaza J. Acta Crystallographica Section A,
        vol. 62, no. 4, 2006, pp. 262–69.


    Parameters
    ----------
    lmax: int
        Maximum multipole mode. (Optional)
        Defaults to sqrt(len(flm)).
    lmin: int
        Minimum multipole mode to compute. (Optional, default 0)
    dmat: DeltaMatrix
        Initialize using the data in another DeltaMatrix to start.
        This will speed up computation for larger el, if dmat.lmax <= lmax.
    dtype: type
        Provide a dtype for the array.
        This defaults to 32-bit floats, which will be good to about four decimal
        places.
    """

    def __init__(self, lmax, lmin=0, dmat=None, dtype=np.float32):
        arrsize = self.estimate_array_size(lmin, lmax)
        self._arr = np.empty(arrsize, dtype=dtype)
        self.dtype = dtype
        self.lmax = lmax
        self.lmin = lmin
        self.size = arrsize
        self._eval(dmat)

    def _eval(self, old_dmat=None):
        arr0 = None
        lstart = 0
        if old_dmat is not None:
            # Start using data from old matrix.
            oln, olx = old_dmat.lmin, old_dmat.lmax
            ln, lx = self.lmin, self.lmax
            # Case 0:  [ ( ) ]
            if (oln <= ln) and (lx <= olx):
                arr0 = old_dmat._arr[tri_base(ln) - tri_base(oln):tri_base(lx + 1) - tri_base(oln)]
                lstart = lx
            # Case 1:  [ ( ] )
            elif (oln <= ln <= olx) and (lx > olx):
                arr0 = old_dmat._arr[tri_base(ln) - tri_base(oln):tri_base(olx + 1) - tri_base(oln)]
                lstart = olx
            # Case 2:  [ ] ( )
            elif (olx < ln):
                base = tri_base(olx) - tri_base(oln)
                arr0 = old_dmat._arr[base: base + el_block_size(olx)]
                lstart = olx    # Start from the topmost block.
        _dmat_eval(self.lmax, self._arr, lstart=lstart, arr0=arr0, lmin=self.lmin)

    @classmethod
    def estimate_array_size(cls, lmin, lmax):
        """Estimate size of the flattened array needed to store Delta_lmm values."""
        return (1 + lmax - lmin) * (6 + 5 * lmax + lmax**2 + 4
                                    * lmin + lmax * lmin + lmin**2) // 6

    @classmethod
    def _get_array_params(cls, lmin=None, lmax=None, arrsize=None, strict=False):
        # Fill in the missing parameter.
        # Only one input may be None at a time!

        if arrsize is None:
            arrsize = cls.estimate_array_size(lmin, lmax)

        if lmax is None:
            lmax = lmin
            while True:
                s = cls.estimate_array_size(lmin, lmax)
                if s > arrsize and strict:
                    raise ValueError("Invalid combination.")
                if s > arrsize:
                    lmax -= 1
                    break
                elif s == arrsize:
                    break
                lmax += 1

        if lmin is None:
            lmin = lmax
            while True:
                s = cls.estimate_array_size(lmin, lmax)
                if s > arrsize and strict:
                    raise ValueError("Invalid combination.")
                if s == arrsize:
                    break
                if s > arrsize:
                    lmin += 1
                    break
                lmin -= 1
        arrsize = cls.estimate_array_size(lmin, lmax)
        return (lmin, lmax, arrsize)

    def __getitem__(self, index):
        """
        Access stored elements, or use symmetry relations for non-stored elements.

        Parameters
        ----------
        index: tuple
            (l, m1, m2) delta matrix entry

        Returns
        -------
        float
            Value of Delta[el, m1, m2]
        """
        (l, m1, m2) = index
        if l < self.lmin:
            raise ValueError("l < lmin. Need to re-evaluate delta matrix.")
        if l > self.lmax:
            raise ValueError("l > lmax. Need to re-evaluate delta matrix.")
        return _access_element(l, m1, m2, self._arr, self.lmin)


@jit(nopython=True)
def _get_matrix_elements(el, m1, m2, arr, outarr):
    """
    Get an array of Delta[el, mp, m1] * Delta[el, mp, m2].

    Results are written into outarr
    """
    for mi, mp in enumerate(range(0, el + 1)):
        outarr[mi] = _access_element(
            el, mp, m1, arr) * _access_element(el, mp, m2, arr)


class HarmonicFunction:
    """
    Methods for calculating Wigner-d functions and spin-weighted spherical harmonics.

    Caches a :class:`DeltaMatrix` to speed up subsequent calculations.

    Not to be instantiated.
    """

    current_dmat = None
    cache_mem_limit = 500 * 2**20   # 500 MiB, lmax ~ 921 for lmin = 0

    # This is the maximum el mode that whose full m1/m2 block can be
    # contained in the allowed memory limit.
    # Reset by set_cache_mem_limit()
    maximum_el = 10238

    def __init__(self):
        raise Exception("HarmonicFunction class is not instantiable.")

    @classmethod
    def set_cache_mem_limit(cls, maxmem):
        newlim = maxmem * 2**20
        if cls.current_dmat is not None and cls.current_dmat._arr.nbytes > newlim:
            raise ValueError(f"Cached DeltaMatrix exceeds new limit {maxmem} MiB."
                             " Clear it using clear_cached_dmat() first.")
        cls.cache_mem_limit = newlim
        max_block_size = cls._est_arrsize_limit(cls.cache_mem_limit)
        cls.maximum_el = np.floor((np.sqrt(1 + 8 * max_block_size) - 3) / 2)

    @classmethod
    def get_cache_details(cls):
        return_dict = {
            'cache_mem_limit': cls.cache_mem_limit,
            'maximum_el': cls.maximum_el
        }
        if cls.current_dmat is not None:
            return_dict.update({
                'lmin': cls.current_dmat.lmin,
                'lmax': cls.current_dmat.lmax,
                'size': cls.current_dmat.size,
            })
        return return_dict

    @classmethod
    def _est_arrsize_limit(cls, maxmem, double=False):
        # maxmem = memory limit in bytes
        dtype = np.float32
        if double:
            dtype = np.float64
        return int(np.floor(maxmem / np.zeros(1, dtype=dtype).nbytes))

    @classmethod
    def _limit_lmin_lmax(cls, lmin, lmax, high, dtype=None):
        """
        Choose new lmin/lmax respecting array size limits.

        Parameters
        ----------
        lmin, lmax: int
            Desired lmin/lmax.
        high: bool
            If the given limits are not possible under memory
            restrictions, this chooses whether the returned range
            should include lmin (False) or lmax (True).
        """
        if dtype is np.float64:
            double = True
        else:
            double = False
        mem_lim = getattr(cls, 'cache_mem_limit')
        max_arrsize = cls._est_arrsize_limit(mem_lim, double=double)
        req_arrsize = DeltaMatrix.estimate_array_size(lmin, lmax)
        limited = req_arrsize > max_arrsize
        if limited and high:
            lmin, _, _ = DeltaMatrix._get_array_params(lmax=lmax, arrsize=max_arrsize)
        elif limited:
            _, lmax, _ = DeltaMatrix._get_array_params(lmin=lmin, arrsize=max_arrsize)

        return lmin, lmax

    @classmethod
    def _set_wigner(cls, lmin, lmax, high=True, dtype=None):
        """
        Cache a DeltaMatrix, respecting memory limits.

        If a DeltaMatrix is already cached, it will use that as a starting point.

        Will recompute everything if the data type changes.
        """
        if lmax > cls.maximum_el:
            raise ValueError("Cannot construct DeltaMatrix within given memory limits.")
        dtype_set = (dtype is not None)
        lmin, lmax = cls._limit_lmin_lmax(lmin, lmax, high=high, dtype=dtype)
        if (cls.current_dmat is None or (dtype_set and cls.current_dmat.dtype != dtype)):
            cls.current_dmat = DeltaMatrix(lmax, lmin=lmin, dtype=dtype)
        elif (cls.current_dmat.lmax < lmax) or (cls.current_dmat.lmin > lmin):
            cls.current_dmat = DeltaMatrix(lmax, lmin=lmin, dmat=cls.current_dmat, dtype=dtype)

    @classmethod
    def wigner_d(cls, el, m1, m2, theta, lmax=None, lmin=None, double_prec=None):
        theta = np.atleast_1d(theta)
        if lmin is None:
            lmin = 0
        if lmax is None:
            lmax = el

        if el < lmin:
            raise ValueError(f"el {el} is less than lmin {lmin}")

        dtype = None
        if double_prec is not None:
            if double_prec:
                dtype = np.float64
            else:
                dtype = np.float32
        cls._set_wigner(lmin, lmax, dtype=dtype)

        mp = np.arange(1, el + 1)
        exp_fac = np.exp(1j * mp[None, :] * theta[..., None])
        dmats = np.empty(el + 1, dtype=float)
        _get_matrix_elements(el, m1, m2, cls.current_dmat._arr, dmats)

        val = (1j) ** (m2 - m1) * (
            np.sum((exp_fac + (-1.)**(m1 + m2 - 2 * el) / exp_fac)
                   * dmats[1:], axis=-1)
            + dmats[0]
        )
        if val.size == 1:
            return complex(val)
        return val.squeeze()

    @classmethod
    def spin_spherical_harmonic(cls, spin, el, em, theta, phi,
                                lmin=None, lmax=None, double_prec=None):
        theta = np.asarray(theta)
        phi = np.asarray(phi)

        if not theta.shape == phi.shape:
            raise ValueError("theta and phi must have the same shape.")

        return (-1.)**(em) * np.sqrt((2 * el + 1) / (4 * np.pi))\
            * np.exp(1j * em * phi) * cls.wigner_d(el, em, -1 * spin,
                                                   theta, lmin=lmin, lmax=lmax,
                                                   double_prec=double_prec)

    @classmethod
    def clear_dmat(cls):
        """Delete cached DeltaMatrix."""
        cls.current_dmat = None


def wigner_d(el, m1, m2, theta, lmin=None, lmax=None, double_prec=None):
    """
    Evaluate the Wigner-d function (little-d) using cached values at pi/2.

    d^l_{m1, m2}(theta)

    Parameters
    ----------
    el, m1, m2: ints
        Indices of the d-function.
    theta: float or ndarray of float
        Angle argument(s) in radians.
    lmin: int
        Precompute the cached DeltaMatrix from some minimum el.
        (Optional. See notes.)
    lmax: int
        Precompute the cached DeltaMatrix up to some maximum el.
        (Optional. See notes.)
    double_prec: bool or None
        Compute the DeltaMatrix using doubles instead of singles.
        If set, this will enforce that the cached DeltaMatrix is in
        either single precision (False) or double (True).
        Otherwise, the cached DeltaMatrix will be kept at whatever
        precision it had previously, or will be made in single precision.
        Default None

    Returns
    -------
    complex or ndarray of complex
        Value of Wigner-d function.
        If multiple theta values are given, multiple values are returned.

    Notes
    -----
    Uses eqn 8 of McEwan and Wiaux (2011), which cites:
        A. F. Nikiforov and V. B. Uvarov (1991)

    When lmin/lmax are not set, the default behavior is to cache a DeltaMatrix
    covering 0 to about el in el.

    If larger el are required, the cached matrix is expanded.

    The total matrix size is limited so the cache doesn't exceed 200 MiB.
    This limit can be changed.
    """
    return HarmonicFunction.wigner_d(
        el, m1, m2, theta, lmin=lmin, lmax=lmax, double_prec=double_prec
    )


def spin_spherical_harmonic(spin, el, em, theta, phi, lmin=None, lmax=None, double_prec=None):
    """
    Evaluate the spin-weighted spherical harmonic.

    Obeys the standard numpy broadcasting for theta and phi.

    Parameters
    ----------
    spin: int
        Spin of the function.
    el, em: ints
        Spherical harmonic mode.
    theta: ndarray or float
        Colatitude(s) to evaluate to, in radians.
    phi: ndarray or float
        Azimuths to evaluate to, in radians.
    lmin: int
        Precompute the cached DeltaMatrix from some minimum el.
        (Optional. See notes on wigner_d.)
    lmax: int
        Precompute the cached Delta matrix up to some maximum el.
        (Optional. See notes on wigner_d.)
    double_prec: bool or None
        Compute the DeltaMatrix using doubles instead of singles.
        (See docstring for wigner_d for details).
        Default None

    Returns
    -------
    complex or ndarray of complex
        Values of the sYlm spin-weighted spherical harmonic function
        at spherical positions theta, phi.

    Notes
    -----
    Uses eqns (2) and (7) of McEwan and Wiaux (2011).
    If theta/phi are arrays, they must have the same shape.
    """
    return HarmonicFunction.spin_spherical_harmonic(
        spin, el, em, theta, phi, lmin=lmin, lmax=lmax, double_prec=double_prec
    )


def get_cached_dmat():
    """Return currently cached DeltaMatrix."""
    return HarmonicFunction.current_dmat


def clear_cached_dmat():
    """Delete cached DeltaMatrix."""
    HarmonicFunction.clear_dmat()


def set_cache_mem_limit(maxmem):
    """
    Set the memory limit for cached DeltaMatrix values.

    Defaults to 200 MiB.

    Also estimates the maximum el that can be contained in this limit by itself.

    Parameters
    ----------
    maxmem: float
        Memory usage limit in MiB for cached DeltaMatrix array.
        Defaults to 200, corresponding with about lmax ~ 540 for lmin = 0.
    """
    HarmonicFunction.set_cache_mem_limit(maxmem)


def get_cache_details():
    """
    Details of the cached DeltaMatrix.

    Returns
    -------
    dict:
        A dictionary containing:
            * cache_mem_limit
                Limit in MiB of the cached DeltaMatrix
            * maximum_el
                Maximum el mode that can be supported by itself
                in the cached DeltaMatrix, under the current limit.

                This is not the same as lmax. Each el mode "block"
                in the matrix consists of (el + 1) * (el + 2) / 2 floats.
                This is the maximum el whose block can fit in the memory limit.
        The following will be included only if a DeltaMatrix is cached:
            * lmin
                Minimum el mode in the cached DeltaMatrix
            * lmax
                Maximum el mode in the cached DeltaMatrix
            * size
                Number of floats in the cached DeltaMatrix
    """
    return HarmonicFunction.get_cache_details()


def _fac(val):
    return factorial(val, exact=True)


def spin_spharm_goldberg(spin, el, em, theta, phi):
    """
    Spin-s spherical harmonic function from Goldberg et al. (1967).

    Parameters
    ----------
    spin: int
        Spin of the function.
    el, em: ints
        Spherical harmonic mode.
    theta: array_like or float
        Colatitude(s) to evaluate to, in radians.
    phi: array_like or float
        Azimuths to evaluate to, in radians.

    Returns
    -------
    complex or ndarray of complex
        Values of the sYlm spin-weighted spherical harmonic function
        at spherical positions theta, phi.

    Notes
    -----
    If theta/phi are arrays, they must have the same shape.

    For nonzero spin, this function is unstable when theta/2 is close to a multiple of pi.

    Also at risk of an integer overflow for el > 10.
    """
    theta = np.asarray(theta)
    phi = np.asarray(phi)

    if not theta.shape == phi.shape:
        raise ValueError("theta and phi must have the same shape.")

    if (spin > 0) and (el < spin):
        return np.zeros_like(theta)

    term0 = (-1.)**em * np.sqrt(
        _fac(el + em) * _fac(el - em) * (2 * el + 1)
        / (4 * np.pi * _fac(el + spin) * _fac(el - spin))
    )
    term1 = np.sin(theta / 2)**(2 * el)

    # This will include divide by zeros. They are removed later.
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='divide by zero encountered')
        warnings.filterwarnings('ignore', message='invalid value encountered in multiply')
        term2 = np.sum([
            binom(el - spin, r) * binom(el + spin, r + spin - em)
            * (-1)**(el - r - spin)
            * np.exp(1j * em * phi) * (1 / np.tan(theta / 2))**(2 * r + spin - em)
            for r in range(el - spin + 1)
        ], axis=0)

        res = term0 * term1 * term2

    if np.isscalar(res) and (np.isclose((theta / 2) % np.pi, 0)):
        res = 0.0
    else:
        res[np.isclose((theta / 2) % np.pi, 0)] = 0.0    # Singularities

    if res.size == 1:
        return complex(res)

    return res



import numpy as np
import pytest

import pyspherical as pysh

"""
This will only run if NUMBA_DISABLE_JIT is in the environmental variables.
For automated testing, this needs to be run as a separate github action.

This allows coverage to see the contents of JIT functions.
"""

pytestmark = pytest.mark.skipif('not "NUMBA_DISABLE_JIT" in os.environ')


def test_delta_eval():
    # _dmat_eval
    # Check that values are equivalent for single and double.

    lmax = 20
    arr0 = np.empty(pysh.DeltaMatrix.estimate_array_size(0, lmax), dtype=np.float32)
    arr1 = np.empty(pysh.DeltaMatrix.estimate_array_size(0, lmax), dtype=np.float64)
    pysh.wigner._dmat_eval(lmax, arr0, lmin=0, lstart=None, arr0=None)

    pysh.wigner._dmat_eval(lmax, arr1, lmin=0, lstart=None, arr0=None)

    # Starting from an lmin
    lmin = 5
    arr2 = np.empty(pysh.DeltaMatrix.estimate_array_size(lmin, lmax), dtype=np.float32)
    base = pysh.utils.tri_base(lmin)
    bsize = pysh.utils.el_block_size(lmin)
    pysh.wigner._dmat_eval(lmax, arr2, lmin=lmin, lstart=lmin, arr0=arr1[base:base + bsize])

    assert np.allclose(arr0, arr1, atol=1e-4)
    assert np.allclose(arr1[pysh.utils.tri_base(lmin):], arr2)


@pytest.mark.parametrize('lmin', [0, 3, 7])
def test_indexing_and_sizes(lmin):

    lmax = 10
    tot = pysh.utils.tri_base(lmin)
    for el in range(lmin, lmax + 1):
        assert pysh.utils.tri_base(el) == tot
        tot += pysh.utils.el_block_size(el)
    assert pysh.DeltaMatrix.estimate_array_size(lmin, lmax) == tot - pysh.utils.tri_base(lmin)

    # Error for invalid index
    with pytest.raises(ValueError, match="Invalid indices"):
        pysh.utils.tri_ravel(0, 1, 0)

    # Check that tri_ravel indices cover the full array.
    inds = []
    for el in range(lmin, lmax + 1):
        for m1 in range(0, el + 1):
            for m2 in range(0, m1 + 1):
                inds.append(pysh.utils.tri_ravel(el, m1, m2))

    expected_size = pysh.DeltaMatrix.estimate_array_size(lmin, lmax)
    assert np.all(sorted(inds) == np.arange(expected_size) + pysh.utils.tri_base(lmin))


def test_ravel_lm():
    # Check that ravel_lm is consistent with unravel_lm:
    lmax = 10
    inds = np.arange(lmax**2)
    el, em = np.array([pysh.utils.ravel_lm(inds[ii]) for ii in range(lmax**2)]).T
    assert np.all(np.abs(em) <= el)
    assert np.all([inds[ii] == pysh.utils.unravel_lm(el[ii], em[ii]) for ii in range(lmax**2)])


def test_access_element():
    # Results of _access_element respect the same symmetries as checked in test_wigner.py
    lmax = 20
    arr0 = np.empty(pysh.DeltaMatrix.estimate_array_size(0, lmax), dtype=np.float32)
    arr2 = np.ones_like(arr0) * np.nan
    pysh.wigner._dmat_eval(lmax, arr0, lmin=0, lstart=None, arr0=None)

    _acc = pysh.wigner._access_element

    with pytest.raises(ValueError, match="Invalid entry in dmat"):
        _acc(2, 0, 0, arr2)

    for ll in range(lmax):
        for m in range(-ll, ll + 1):
            for mm in range(-ll, ll + 1):
                val1 = _acc(ll, m, mm, arr0)

                assert np.isclose(val1, (-1)**(m - mm) * _acc(ll, -m, -mm, arr0))
                assert np.isclose(val1, (-1)**(m - mm) * _acc(ll, mm, m, arr0))
                assert np.isclose(val1, _acc(ll, -mm, -m, arr0))
                assert np.isclose(
                    _acc(ll, m, mm, arr0), (-1)**(ll - mm) * _acc(ll, -m, mm, arr0), atol=1e-10)


def test_transform():
    # To cover the compiled methods in transforms.py
    lmax = 10
    theta, phi = pysh.utils.get_grid_sampling(lmax=lmax)

    amp = 20
    dat = np.ones((theta.size, phi.size)) * amp

    flm = pysh.forward_transform(dat, phi, theta, lmax)
    res = pysh.inverse_transform(flm, phi, theta, lmax)
    assert np.isclose(flm[0], amp * np.sqrt(4 * np.pi))
    assert np.allclose(res, amp, atol=1e-4)


def test_wigner_d_matrix_el():
    # Check the _get_matrix_elements function

    dmat = pysh.DeltaMatrix(6)

    el = 5
    m1 = 3
    m2 = 2

    test = []
    for mp in range(0, el + 1):
        test.append(dmat[el, mp, m1] * dmat[el, mp, m2])

    test = np.array(test)

    comp = np.empty(test.size, dtype=np.float32)
    pysh.wigner._get_matrix_elements(el, m1, m2, dmat._arr, comp)

    assert np.allclose(comp, test)
